from .mirror import Mirror
from .parabola import Parabola
from .camera import Camera


def initialise_element(el):
    match el['type']:
        case 'mirror':
            return Mirror.from_dict(el)
        case 'parabola':
            return Parabola.from_dict(el)
        case 'camera':
            return Camera.from_dict(el)
        case _ as type:
            print(f'Unknown element type {type}')
            sys.exit(1)



from . import Beam
class BeamFocusing(Beam):
    e_field_at_infinity = None
    e_r_interpolant = None
    e_i_interpolant = None
    beam_waist = 0.
    rayleigh = 0.
    u_norm = None
    v_norm = None

    def __init__(self, position, normal, wavenumber, beam_width, u, v, focal_length, e_field, do_prop=True):
        '''
        For this beam type, `position` should be the position of the focusing optic
        The focal point is located at `position + focal_length * normal`
        beam_width is defined at the focusing optic, not at the focal point. For a
        Gaussian beam profile it should be equal to the 1/e field radius of the beam.
        u, v are also defined at the focusing optic.
        e_field is taken to be equal to the E field at infinity up to a factor w0/w(z),
        when expressed as a function of (u, v)/w(z), which is true for focal_length >> z_R
        Therefore you should make sure that focal_length >> z_R if you want this to be accurate
        '''

        super().__init__(position, normal, wavenumber)
        self.e_field_at_infinity = e_field
        F0 = focal_length / (2 * beam_width)
        # F_actual = F0 * ( (1/2 (k0 W / 4F0)^2) - sqrt( (1/2 (k0 W / 4F0)^2)^2 - (k0 W / 4F0)^2 ) )
        self.beam_waist = 4 * F0 / self.wavenumber
        self.rayleigh = 8 * F0**2 / self.wavenumber
        self.beam_z = -focal_length
        self.u_norm = np.sqrt(2) * u / beam_width
        self.v_norm = np.sqrt(2) * v / beam_width
        if do_prop:
            self.propagate_z(0)

    def copy_with(self, position, normal):
        B = BeamFocusing(position, normal, self.wavenumber,
                1, 0, 0, -self.beam_z, self.e_field_at_infinity.copy(), False)
        B.beam_waist = self.beam_waist
        B.rayleigh = self.rayleigh
        B.u_norm = self.u_norm.copy()
        B.v_norm = self.v_norm.copy()
        B.e_r_interpolant = self.e_r_interpolant
        B.e_i_interpolant = self.e_i_interpolant
        return B

    def width(self):
        return self.beam_waist * np.sqrt(1 + (self.beam_z/self.rayleigh)**2)

    def max_width(self):
        return self.beam_waist * np.sqrt(1 + (self.beam_z/self.rayleigh)**2)


    def e_field(self, u, v):
        beam_width = self.width()
        u_norm = np.sqrt(2) * u / beam_width
        v_norm = np.sqrt(2) * v / beam_width

        if u.ndim==1:
            e_norm = (
                self.e_r_interpolant(u_norm, v_norm)
                + 1j * self.e_i_interpolant(u_norm, v_norm)
            ).T
        else:
            u_full = np.linspace(u_norm.min(), u_norm.max(), u.shape[0]*2)
            v_full = np.linspace(v_norm.min(), v_norm.max(), v.shape[1]*2)
            e_full = (
                self.e_r_interpolant(u_full, v_full)
                + 1j * self.e_i_interpolant(u_full, v_full)
            ).T
            actual_terp = RegularGridInterpolator((u_full, v_full), e_full, bounds_error=False, fill_value=0)
            e_norm = actual_terp((u_norm, v_norm))

        # TODO Add other phase factors here, e.g. kz & the parabolic phase
        # Maybe also guoy, but I think that's handled in the FrFT already.
        return e_norm / np.sqrt(1 + (self.beam_z/self.rayleigh)**2)

    def propagate_z(self, z):
        super().propagate_z(z)

        tan_psi = self.beam_z / self.rayleigh

        ku = util.k_from_x(self.u_norm)
        kv = util.k_from_x(self.v_norm)
        u_prime = ku / np.sqrt(1 + tan_psi**2)
        v_prime = kv / np.sqrt(1 + tan_psi**2)

        e_field = FrFT_tangent(
            self.v_norm, v_prime, FrFT_tangent(
                self.u_norm, u_prime, self.e_field_at_infinity.T, tan_psi
            ).T, tan_psi
        )
        self.e_r_interpolant = interp2d(u_prime, v_prime, e_field.real.T, bounds_error=False, fill_value=0)
        self.e_i_interpolant = interp2d(u_prime, v_prime, e_field.imag.T, bounds_error=False, fill_value=0)




import numpy as np
from numpy.fft import fft2, ifft2, ifftshift

from scipy.interpolate import RectBivariateSpline

from . import Beam

class BeamGouyFresnel(Beam):
    def __init__(self, position, normal, wavenumber, beam_width, u, v, focal_length, e_field):
        '''
        For this beam type, `position` should be the position of the focusing optic
        The focal point is located at `position + focal_length * normal`
        beam_width is defined at the focusing optic, not at the focal point. For a
        Gaussian beam profile it should be equal to the 1/e field radius of the beam.
        u, v are also defined at the focusing optic.
        e_field is taken to be equal to the E field at infinity up to a factor w0/w(z),
        when expressed as a function of (u, v)/w(z), which is true for focal_length >> z_R
        Therefore you should make sure that focal_length >> z_R if you want this to be accurate
        '''

        super().__init__(position, normal, wavenumber)
        self.apodise = None
        self.e_field_arr = e_field
        F0 = focal_length / (2 * beam_width)
        self.beam_waist = 4 * F0 / self.wavenumber
        self.rayleigh = 8 * F0**2 / self.wavenumber
        self.beam_z = -focal_length
        self.psi = np.arctan(self.beam_z / self.rayleigh)
        self.u_norm = np.sqrt(2) * u / beam_width
        self.v_norm = np.sqrt(2) * v / beam_width
        self.max_step = np.pi / 10

        dx = np.diff(self.u_norm, axis=0).mean()
        dy = np.diff(self.v_norm, axis=1).mean()

        self.r2 = self.u_norm**2 + self.v_norm**2

        kx = 2*np.pi * self.u_norm / (self.u_norm.shape[0]*dx**2)
        ky = 2*np.pi * self.v_norm / (self.v_norm.shape[1]*dy**2)
        self.k2 = kx**2 + ky**2

    def copy_with(self, position, normal):
        beam_width = self.width()
        return BeamGouyFresnel(position, normal, self.wavenumber,
                beam_width, self.u_norm * beam_width / np.sqrt(2), self.v_norm * beam_width / np.sqrt(2), -self.beam_z, self.e_field_arr.copy())

    def width(self):
        return self.beam_waist * np.sqrt(1 + (self.beam_z/self.rayleigh)**2)

    def max_width(self):
        return self.beam_waist * np.sqrt(1 + (self.beam_z/self.rayleigh)**2)


    def e_field(self, u, v):
        beam_width = self.width()
        u_norm = np.sqrt(2) * u / beam_width
        v_norm = np.sqrt(2) * v / beam_width

        e_norm = np.zeros((u_norm.shape[0], v_norm.shape[1]), 'complex')
        valid = (self.u_norm.min() < u_norm) & (u_norm < self.u_norm.max())
        valid &= (self.v_norm.min() < v_norm) & (v_norm < self.v_norm.max())

        e_norm[valid] = self.e_r_terp.ev(u_norm[valid], v_norm[valid]) + 1j * self.e_i_terp.ev(u_norm[valid], v_norm[valid])
        # TODO Add other phase factors here, e.g. kz & the parabolic phase
        # Maybe also guoy, but I think that's handled in the FrFT already.
        return e_norm / np.sqrt(1 + (self.beam_z/self.rayleigh)**2)

    def propagate_z(self, dz):
        super().propagate_z(dz)

        new_psi = np.arctan(self.beam_z / self.rayleigh)
        Dpsi = new_psi - self.psi
        N_steps = np.ceil(Dpsi / self.max_step).astype('int')

        dpsi = Dpsi / N_steps

        # Free space propagation factor
        spectral_phase_halfstep = ifftshift(np.exp(-1j * self.k2/4 * dpsi/2))
        spectral_phase = ifftshift(np.exp(-1j * self.k2/4 * dpsi))

        # pseudo-ior function
        spatial_phase = np.exp(-1j * self.r2 * dpsi)

        Eff = fft2(self.e_field_arr)
        E = ifft2(Eff * spectral_phase_halfstep)

        for _ in range(N_steps-1):
            E *= spatial_phase
            if self.apodise is not None:
                E *= self.apodise

            Eff = fft2(E)
            E = ifft2(Eff * spectral_phase)

        E *= spatial_phase
        Eff = fft2(E)
        self.e_field_arr = ifft2(Eff * spectral_phase_halfstep)
        if self.apodise is not None:
            self.e_field_arr *= self.apodise
        
        self.e_r_terp = RectBivariateSpline(self.u_norm[:,0], self.v_norm[0,:], self.e_field_arr.real)
        self.e_i_terp = RectBivariateSpline(self.u_norm[:,0], self.v_norm[0,:], self.e_field_arr.imag)




from contextlib import contextmanager

import numpy as np
from numpy import array as arr
from numpy import sqrt, sin, cos, arctan2, dot, outer

import time

@contextmanager
def timer(message):
    print(message)
    spaces = len(message) - len(message.lstrip(' '))
    t0 = time.monotonic_ns()
    try:
        yield
    finally:
        t1 = time.monotonic_ns()
        print(spaces * ' ' + f'Took {(t1-t0)*1e-6:0.3f}ms')


def normal(theta, phi):
    '''
    Returns a normal vector for a given pitch theta and yaw phi
    '''
    return arr([
        cos(theta) * sin(phi),
        sin(theta),
        cos(theta) * cos(phi)
    ])


def angles(n):
    '''
    Returns the pitch and yaw angle of a given normal vector
    '''
    x, y, z = n
    phi_beam = arctan2(x, z)
    theta_beam = arctan2(y, sqrt(x**2 + z**2))
    return theta_beam, phi_beam


def rot_matrix_angles(theta, phi):
    '''
    This matrix rotates from object coordinates to world coordinates
    Its columns correspond respectively to
        * The object u vector (in-plane, horizontal)
        * The object v vector (in-plane, with a vertical component)
        * The object w (normal) vector
    Multiply an object uvw vector by this matrix
    to transform to world space
    Multiply a world-space xyz vector by its transpose
    to transform to object coordinates.
    '''
    return arr(
        [
            [cos(phi), -sin(theta)*sin(phi), cos(theta)*sin(phi)],
            [0, cos(theta), sin(theta)],
            [-sin(phi), -sin(theta)*cos(phi), cos(theta)*cos(phi)]
        ]
    )


def rot_matrix_normal(object_normal):
    '''
    This matrix rotates from object coordinates to world coordinates
    Its columns correspond respectively to
        * The object u vector (in-plane, horizontal)
        * The object v vector (in-plane, with a vertical component)
        * The object w (normal) vector
    Multiply an object uvw vector by this matrix
    to transform to world space.
    Multiply a world-space xyz vector by its transpose
    to transform to object coordinates.
    '''
    x, y, z = object_normal
    scale = sqrt(1 - y**2)
    u_vec = arr([z, 0, -x]) / scale
    v_vec = arr([-x * y, x**2 + z**2, -y * z]) / scale
    return arr([u_vec, v_vec, object_normal]).T


def projection_matrix(n_1, n_2):
    '''
    This matrix projects a vector along n1 such that
    the result's inner product with n2 is zero.
    Usually n1 will be a beam normal vector
    and n2 the normal of an optical element
    '''
    return np.eye(3) - outer(n_1, n_2) / dot(n_1, n_2)


def reflection_matrix(n_r):
    '''
    This matrix reflects a vector in the plane whose
    normal is n_r
    '''
    return np.eye(3) - 2 * outer(n_r, n_r) / dot(n_r, n_r)


def centred_fft(x, f, **kwargs):
    F = np.fft.fftshift(
        np.fft.fft(
            np.fft.ifftshift(f), **kwargs
        )
    )
    k_x = k_from_x(x)
    return k_x, F


def centred_ifft(x, f, **kwargs):
    F = np.fft.fftshift(
        np.fft.ifft(
            np.fft.ifftshift(f), **kwargs
        )
    )
    k_x = k_from_x(x)
    return k_x, F


def centred_fft2(x, y, f, **kwargs):
    F = np.fft.fftshift(
        np.fft.fft2(
            np.fft.ifftshift(f), **kwargs
        )
    )
    k_x = k_from_x(x)
    k_y = k_from_x(y)
    return k_x, k_y, F


def centred_ifft2(x, y, f, **kwargs):
    F = np.fft.fftshift(
        np.fft.ifft2(
            np.fft.ifftshift(f), **kwargs
        )
    )
    k_x = k_from_x(x)
    k_y = k_from_x(y)
    return k_x, k_y, F


def k_from_x(x):
    return np.fft.fftshift(np.fft.fftfreq( x.size, np.diff(x).mean())) * 2*np.pi



class Beam:
    position = None
    normal = None
    wavenumber = 0
    twist = 0
    beam_z = 0
    clipping = False

    def __init__(self, position, normal, wavenumber):
        self.position = position
        self.normal = normal
        self.wavenumber = wavenumber

    def copy_with(self, position, normal):
        return Beam(position, normal, self.wavenumber)

    def width(self):
        return 0

    def max_width(self):
        return 3*self.width()

    def e_field(self, x, y):
        raise NotImplementedError()

    def propagate_z(self, z):
        self.beam_z += z
        self.position += z * self.normal

    def clip(self, clip_information):
        pass


class BeamMissed(Beam):
    def __init__(self):
        pass

from .geometric import BeamGaussianGeometric
from .fresnel_gouy import BeamGouyFresnel




import numpy as np
import matplotlib.pyplot as plt

import propagate


def plot_focus(x, y, A, W, focal_distance, wavelength, Nstep):
    '''
    x: vector, shape (n_x, 1) in your units of choice (e.g. cm)
    y: vector, shape (1, n_y) in your units of choice
    A: matrix, shape (n_x, n_y) the complex e-field you want to focus
    W: scalar, the `radius' of A, in the same units as x and y
    focal_distance: the focal length of the optic, same units as w
    lambda: wavelength of the laser.
    Nstep: Number of steps to take in the Guoy phase angle

    If lambda is given in different units to the macroscopic distances
    x,y,W,focal_distance then the results will be plotted in the units of 
    lambda
    '''

    guoy = np.linspace(-np.pi/2, np.pi/2, Nstep+1)
    x0 = x/W
    y0 = y/W
    A_focus = propagate.propagate_to(x0, y0, A, (0,))[0]
    # This yields the x=0 slice (or as close to it as we can get)
    print('now')
    i_xmin = np.argmin(abs(x0))
    A_slice = propagate.propagate_slices(x0, y0, A, guoy, slices=[np.s_[i_xmin, :]])[0]

    w0 = wavelength * focal_distance / W / np.pi
    zR = np.pi * w0**2 / wavelength
    z0 = np.tan(guoy)
    w_of_z = w0 / np.cos(guoy)

    I_focus = abs(A_focus)**2 * (W/w0)**2
    I_slice = abs(A_slice)**2 * (W/w_of_z[:,None])**2 * z0[:,None]**2

    fig, axs = plt.subplots(1, 2)

    axs[0].imshow(I_focus.T, extent=(x0.min() * w0, x0.max() * w0, y0.min() * w0, y0.max() * w0))

    Z0, Y0 = np.meshgrid(z0, x0, indexing='ij')
    axs[1].pcolormesh(Z0 * zR, Y0 * w_of_z[:,None], I_slice, shading='gouraud')
    axs[1].set_xlim([-10*zR, 10*zR])
    axs[1].set_ylim([-10*w0, 10*w0])

    return A_slice



import numpy as np

class Propagator:
    pass

class Hermite(Propagator):
    @staticmethod
    def generate(x,bandlimit=None):
        # This generates Hermite functions ψn(x / sqrt(2))
        # Which are the eigenfunctions of the focusing envelope equation

        dx = np.diff(x.flat).mean()

        if not bandlimit:
            bandlimit = x.size
        elif bandlimit == 'conservative':
            # Bandlimit to keep functions within bounds of x
            bandlimit = np.max(x**2)
        elif bandlimit == 'fourier':
            bandlimit = (np.pi / (2 * dx))**2

        n_func = np.ceil(bandlimit).astype('int')

        V = np.zeros((n_func, x.size), 'longdouble')
        x = x.astype('longdouble')

        V[0,:] = np.exp(700-x**2)
        V[1,:] = np.exp(700-x**2) * 2 * x

        for j in np.arange(1, n_func-1):
            V[j+1, :] = np.sqrt(1 / (j+1)) * (2 * x * V[j, :] - np.sqrt(j) * V[j-1,:])

        return V * (2 / np.pi)**0.25 * np.exp(-700)

    def __init__(self, x, y, bandlimit=None):
        self.x, self.y = x, y
        self.dx = np.diff(x).mean()
        self.dy = np.diff(y).mean()
        if type(bandlimit) is not tuple:
            bandlimit = (bandlimit, bandlimit)
        self.U = Hermite.generate(x, bandlimit[0])
        self.V = Hermite.generate(y, bandlimit[1])
        self.bandlimit = (self.U.shape[0], self.V.shape[0])

        self.mode_numbers = (
                np.arange(self.bandlimit[0])[:, None]
              + np.arange(self.bandlimit[1])[None, :]
        )

    def copy(self):
        ''' TODO: actually copy'''
        return self

    def decompose(self, A):
        return self.U @ A @ self.V.T * self.dx * self.dy

    def get_phases(self, dphi):
        return np.exp(-1j * self.mode_numbers * dphi)

    def recompose(self, coeffs):
        return self.U.T @ coeffs @ self.V


    def propagate_iter(self, A, psi, psi0=-np.pi/2):
        # x: (n_x,) 1d x axis
        # y: (n_y,) 1d y axis
        # A: (n_x, n_y) 2d complex field amplitude. Note 'ij' meshgrid indexing.
        # psi: (n_psi,) points to evaluate A(psi) at

        coeffs = self.decompose(A)

        for i, p in enumerate(psi):
            A_p = self.recompose(coeffs * self.get_phases(p - psi0))
            print(i)
            yield A_p

    def propagate_slices(self, A, psi, slices, psi0=-np.pi/2, bandlimit=None):
        A_all_sliced = [
            np.zeros(psi.shape + A[slc].shape, dtype='complex') for slc in slices
        ]
        for n, A in enumerate(self.propagate_iter(A, psi, psi0, bandlimit)):
            for A_sliced, slc in zip(A_all_sliced, slices):
                A_sliced[n] = A[slc]

        return A_all_sliced

    def propagate_to(self, A, psi, psi0=-np.pi/2, bandlimit=None):
        # x: (n_x,) 1d x axis
        # y: (n_y,) 1d y axis
        # A: (n_x, n_y) 2d complex field amplitude. Note 'ij' meshgrid indexing.
        # psi: (n_psi,) points to evaluate A(psi) at
         
        return tuple(self.propagate_iter(A, psi, psi0, bandlimit))




import numpy as np

from .. import parser
from .. import util


class Camera():
    @staticmethod
    def from_dict(config_dict):
        position = parser.position(config_dict['position'])
        angles = parser.euler_angles(config_dict['pointing'])
        px = np.array(config_dict['pixels'])
        size = np.array(config_dict['sensor_size'])
        magnification = config_dict['magnification']
        return Camera(position, angles, px, size, magnification)

    def __init__(self, position, angles, px, size, magnification):
        self.position = position
        self.pitch = angles[0]
        self.yaw = angles[1]
        self.sensor = np.zeros(px)
        self.size = size
        self.magnification = magnification

    def propagate(self, beam):
        cam_normal = util.normal(self.pitch, self.yaw)
        Uc, Vc, Nc = util.rot_matrix_normal(cam_normal).T
        Ub, Vb, Nb = util.rot_matrix_normal(beam.normal).T

        if Nc @ Nb > 0:
            print('Camera looking in same direction as beam is looking')
            self.sensor[:] = 0
            return beam

        # Place the origin at the camera centre for now
        dpos = beam.position - self.position
        proj = util.projection_matrix(Nb, Nc)
        sensor_intersect = proj @ dpos

        dz = (sensor_intersect - dpos) @ Nb
        if dz < 0:
            print('Beam moving away from camera')
            self.sensor[:] = 0
            return beam

        with util.timer('    Propagating beam to camera plane'):
            beam.propagate_z(dz)

        max_dist = np.sqrt((self.size**2).sum()) + 2*beam.max_width()

        if sensor_intersect @ sensor_intersect > max_dist**2:
            print('Beam missed camera')
            self.sensor[:] = 0
            return beam

        cam_u = np.linspace(-1, 1, self.sensor.shape[0]) * self.size[0]/2 / self.magnification
        cam_v = np.linspace(-1, 1, self.sensor.shape[1]) * self.size[1]/2 / self.magnification
        cam_u_vals, cam_v_vals = np.meshgrid(cam_u, cam_v, indexing='ij')
        cam_x = cam_u_vals * Uc[0] + cam_v_vals * Vc[0] - sensor_intersect[0]
        cam_y = cam_u_vals * Uc[1] + cam_v_vals * Vc[1] - sensor_intersect[1]
        cam_z = cam_u_vals * Uc[2] + cam_v_vals * Vc[2] - sensor_intersect[2]

        proj_to_beam = util.projection_matrix(Nb, Nb)

        beam_x = (proj_to_beam[0, 0] * cam_x +
                  proj_to_beam[0, 1] * cam_y +
                  proj_to_beam[0, 2] * cam_z)

        beam_y = (proj_to_beam[1, 0] * cam_x +
                  proj_to_beam[1, 1] * cam_y +
                  proj_to_beam[1, 2] * cam_z)

        beam_z = (proj_to_beam[2, 0] * cam_x +
                  proj_to_beam[2, 1] * cam_y +
                  proj_to_beam[2, 2] * cam_z)

        beam_u = Ub[0] * beam_x + Ub[1] * beam_y + Ub[2] * beam_z
        beam_v = Vb[0] * beam_x + Vb[1] * beam_y + Vb[2] * beam_z

        self.sensor = abs(beam.e_field(beam_u, beam_v))**2

    def read(self):
        return self.sensor

    def plot(self):
        pass



import numpy as np
from numpy import array as arr
from numpy import sqrt, sin, cos, dot, outer

from ..beams import BeamMissed
from .. import parser
from .. import util


class Mirror:
    @staticmethod
    def from_dict(config_dict):
        position = parser.position(config_dict['position'])
        correct_angles = parser.euler_angles(config_dict['pointing'])
        try:
            radius = config_dict['radius']
        except KeyError:
            radius = float('inf')

        return Mirror(position, correct_angles, radius)

    def __init__(self, position, correct_angles, radius):
        self.position = position
        self.pitch, self.yaw = correct_angles
        self.radius = radius

    def adjust_yaw(self, d_yaw):
        self.yaw += d_yaw

    def adjust_pitch(self, d_pitch):
        self.pitch += d_pitch

    @property
    def normal(self):
        return util.normal(self.pitch, self.yaw)

    def rotation_matrix(self):
        return util.rot_matrix_angles(self.pitch, self.yaw)

    def plot(self):
        '''
        Not yet implemented
        '''
        pass

    def propagate(self, beam):
        if beam.normal.dot(beam.position - self.position) > 0:
            print('travelling away from mirror')
            return BeamMissed()

        mirror_normal = self.normal
        if dot(mirror_normal, beam.normal) > 0:
            print('mirror facing away from beam')
            return BeamMissed()

        proj_beam = util.projection_matrix(beam.normal, mirror_normal)
        reflect_matrix = util.reflection_matrix(mirror_normal)

        hit_location = proj_beam @ (beam.position - self.position)
        new_location = hit_location + self.position
        new_normal = reflect_matrix @ beam.normal

        proj_new = np.eye(3) - outer(new_normal, mirror_normal) / dot(new_normal, mirror_normal)

        if dot(hit_location, hit_location) > self.radius**2:
            print('missed mirror')
            return BeamMissed()

        if beam.clipping:
            # This is as-yet unimplemented
            raise NotImplementedError()
            rot_mirror = self.rotation_matrix()
            rot_beam = util.rot_matrix_normal(beam.normal)
            rot_new = util.rot_matrix_normal(new_normal)
            truncated_identity = np.eye(3)[:, :2]
            m_beam = (rot_mirror @ truncated_identity).T @ proj_beam @ (rot_beam @ truncated_identity)
            m_new = (rot_mirror @ truncated_identity).T @ proj_new @ (rot_new @ truncated_identity)
            twist = np.linalg.inv(m_new) @ m_beam
            twist_angle = arctan2(twist[0,1], twist[0,0])
            beam.twist += twist_angle
            beam.clip()
        ''' Unconverted plotting code
        figure(1)
        plot3([beam_location(3), new_location(3)],...
              [beam_location(1), new_location(1)],...
              [beam_location(2), new_location(2)], 'r')
        axis equal
        xlim([0,1]+[-1,1]*0.1)
        ylim([0,1]+[-1,1]*0.1)
        hit_uvw = rot_mirror' * hit_location
        figure(2)
        subplot( 1, N_mirrors, i)
        cla
        hold on
        scatter( hit_uvw(1), hit_uvw(2) )
        plotMirror3d([0,0,0], mirror_radius, pi/2, 0)
        view(2)
        axis equal
        xlim([-1,1]*mirror_radius)
        ylim([-1,1]*mirror_radius)
        '''
        beam.position = new_location
        beam.normal = new_normal
        return beam




import numpy as np

from . import Beam

class BeamGaussianGeometric(Beam):
    beam_waist = 0
    rayleigh = 0

    def __init__(self, position, normal, wavenumber, beam_waist, distance_to_focus):
        super().__init__(position, normal, wavenumber)
        self.beam_waist = beam_waist
        self.rayleigh = wavenumber * beam_waist**2 / 2
        self.beam_z = -distance_to_focus

    def copy_with(self, position, normal):
        return BeamGaussianGeometric(position, normal, self.wavenumber, self.beam_waist, -self.beam_z)

    def width(self):
        return self.beam_waist * np.sqrt(1 + (self.beam_z/self.rayleigh)**2)

    def e_field(self, x, y):
        beam_width = self.width()
        return np.exp(-(x**2+y**2)/beam_width**2) * np.sqrt(np.pi)/beam_width

class BeamFresnel(Beam):
    e_field_complex = None

    def __init__(self, position, normal, wavenumber, u, v, e_field):
        super().__init__(position, normal, wavenumber)
        self.u = u
        self.v = v
        self.e_field_complex = e_field

    def copy_with(self, position, normal):
        return BeamFresnel(position, normal, self.wavenumber,
                self.u, self.v, self.e_field_complex)

    def width(self):
        # approximation of beam width
        intensity = abs(self.e_field_complex)**2
        u_pos = (self.u[:,None] * intensity).mean() / intensity.mean()
        v_pos = (self.v[None,:] * intensity).mean() / intensity.mean()
        u_var = (((self.u[:,None] - u_pos)**2 * intensity).mean() / intensity.mean())
        v_var = (((self.v[None,:] - v_pos)**2 * intensity).mean() / intensity.mean())
        avg_var = (u_var + v_var)
        return np.sqrt(avg_var)

    def max_width(self):
        return np.sqrt(np.range(self.u)**2 + np.range(self.v)**2) / 2

    def e_field(self, x, y):
        return self.e_field_complex

    def propagate_z(self, z):
        super().propagate_z(z)
        # Use Angular Spectrum Method
        kx, ky, fourier = util.centred_fft2(self.u, self.v, self.e_field_complex)
        kz = np.sqrt(self.wavenumber**2 - (kx**2 + ky**2) + 0j)
        fourier *= np.exp(1j * kz * z)
        self.e_field_complex = util.centred_ifft2(self.u, self.v, fourier)




import numpy as np

from .beams import BeamGaussianGeometric
from . import parser

class BeamInjection:
    @staticmethod
    def from_dict(config_dict):
        position = parser.position(config_dict['position'])
        direction = parser.normal(config_dict['direction'])
        wavelength = config_dict['wavelength']
        beam_type = BeamGaussianGeometric
        beam_waist = config_dict['beam_waist']

        return BeamInjection(
            position,
            direction,
            wavelength,
            beam_type,
            beam_waist=beam_waist,
            distance_to_focus=0
        )

    def __init__(self, position, direction, wavelength,
                 beam_type, **beam_parameters):
        self.position = position
        self.beam_normal = direction
        self.wavenumber = 2*np.pi / wavelength
        self.beam_type = beam_type
        self.beam_parameters = beam_parameters

    def get_beam(self):
        return self.beam_type(
            position=self.position,
            normal=self.beam_normal,
            wavenumber=self.wavenumber,
            **self.beam_parameters
        )



from numpy import pi, sin, cos, arctan2, sqrt, array

def position(entry):
    match entry:
        case [x,y,z]:
            return array([x,y,z])

def angle(entry):
    match entry:
        case {'degrees': value}:
            return pi/180 * value
        case {'radians': value}:
            return value
        case {'cycles': value}:
            return 2*pi * value
        case value:
            return value

def normal(entry):
    named = {'x':(1,0,0), 'y':(0,1,0), 'z':(0,0,1)}
    match entry:
        case {'auto': True}:
            return 'auto'
        case {'axis': name} if name in named:
            return named[name]
        case {'normal': [nx,ny,nz]}:
            total = (nx**2 + ny**2 + nz**2)**0.5
            return array([nx/total, ny/total, nz/total])
        case {'angles': [theta, phi]}:
            return (cos(theta)*sin(phi), sin(theta), cos(theta)*cos(phi))

def euler_angles(entry):
    named = {'x':(0,pi/2), 'y':(pi/2,0), 'z':(0,0)}
    match entry:
        case {'auto': True}:
            return 'auto'
        case {'axis': name} if name in named:
            return named[name]
        case {'normal': [nx,ny,nz]}:
            return (arctan2(ny, sqrt(nx**2 + nz**2)), arctan2(nx, nz))
        case {'angles': [theta, phi]}:
            return (theta, phi)



import numpy as np

from . import Beam
from ..propagators.hermite import Hermite
from .. import util

from scipy.interpolate import interp2d, RegularGridInterpolator

import matplotlib.pyplot as plt

class BeamHermite(Beam):
    '''
    A beam class suitable for both collimated and focusing propagation.
    Models the beam profile by decomposition into Hermite modes.
    '''
    hermite = None
    coeffs = None


    def __init__(self, position, normal, wavenumber, beam_width, u, v, focal_length, e_field, initialise=True):
        '''
        For this beam type, `position` should be the position of the focusing optic
        The focal point is located at `position + focal_length * normal`
        beam_width is defined at the focusing optic, not at the focal point. For a
        Gaussian beam profile it should be equal to the 1/e field radius of the beam.
        u, v are also defined at the focusing optic.
        e_field is taken to be equal to the E field at infinity up to a factor w0/w(z),
        when expressed as a function of (u, v)/w(z), which is true for focal_length >> z_R
        Therefore you should make sure that focal_length >> z_R if you want this to be accurate
        '''

        super().__init__(position, normal, wavenumber)
        if initialise:
            self.e_field_current = e_field
            F0 = focal_length / (2 * beam_width)
            # F_actual = F0 * ( (1/2 (k0 W / 4F0)^2) - sqrt( (1/2 (k0 W / 4F0)^2)^2 - (k0 W / 4F0)^2 ) )
            self.beam_waist = 4 * F0 / self.wavenumber
            self.rayleigh = 8 * F0**2 / self.wavenumber
            self.beam_z = -focal_length
            self.u_norm = u / beam_width
            self.v_norm = v / beam_width
            self.psi_0 = np.arctan(self.beam_z / self.rayleigh)
            self.hermite = Hermite(self.u_norm, self.v_norm, 900)
            field_factor = np.sqrt(1 + (self.beam_z / self.rayleigh)**2)
            self.coeffs = self.hermite.decompose(e_field * field_factor)

    def setup(self, waist, rayleigh, beam_z, u_norm, v_norm, psi_0, hermite, coeffs, e_field):
            self.beam_waist = waist
            self.rayleigh = rayleigh
            self.beam_z = beam_z
            self.u_norm = u_norm.copy()
            self.v_norm = v_norm.copy()
            self.psi_0 = psi_0
            self.hermite = hermite.copy()
            self.coeffs = coeffs.copy()
            self.e_field_current = e_field.copy()


    def copy_with(self, position, normal):
        B = BeamHermite(position, normal, self.wavenumber,
            None, None, None, None, None, False)
        B.setup(
            self.beam_waist, self.rayleigh, self.beam_z, self.u_norm,
            self.v_norm, self.psi_0, self.hermite, self.coeffs, self.e_field_current
        )
        return B

    def width(self):
        return self.beam_waist * np.sqrt(1 + (self.beam_z/self.rayleigh)**2)

    def max_width(self):
        return self.beam_waist * np.sqrt(1 + (self.beam_z/self.rayleigh)**2)

    def e_field(self, u, v):
        beam_width = self.width()
        u_norm = u / beam_width
        v_norm = v / beam_width

        u_full = np.linspace(u_norm.min(), u_norm.max(), u.shape[0]*4)
        v_full = np.linspace(v_norm.min(), v_norm.max(), v.shape[1]*4)

        H_reconstruct = Hermite(u_full, v_full, self.hermite.bandlimit)

        psi_new = np.arctan(self.beam_z / self.rayleigh)
        self.coeffs[:] = 0
        self.coeffs[-1,-1] = 1
        e_full = H_reconstruct.recompose(self.coeffs * self.hermite.get_phases(psi_new - self.psi_0))
        actual_terp = RegularGridInterpolator((u_full, v_full), e_full, bounds_error=False, fill_value=0)
        e_norm = actual_terp((u_norm, v_norm))
        
        #fig, axs = plt.subplots(1,2)
        #axs[0].imshow(H_reconstruct.U)
        #axs[1].imshow(H_reconstruct.V)
        #plt.show()

        return e_norm / np.sqrt(1 + (self.beam_z / self.rayleigh)**2)

    def propagate_z(self, z):
        super().propagate_z(z)

        psi_new = np.arctan(self.beam_z / self.rayleigh)

        self.e_field_current = self.hermite.recompose(self.coeffs * self.hermite.get_phases(psi_new - self.psi_0))

        self.e_r_interpolant = interp2d(self.u_norm, self.v_norm, self.e_field_current.T.real, bounds_error=False, fill_value=0)
        self.e_i_interpolant = interp2d(self.u_norm, self.v_norm, self.e_field_current.T.imag, bounds_error=False, fill_value=0)



from .util import timer

def propagate(beam_injector, optical_system):
    ''' Unconverted plot code
    figure(1);
    cla;
    hold on;
    scatter3(mirror_locations(:,3), mirror_locations(:,1), mirror_locations(:,2), 50, 'ob');
    plot3(mirror_locations(:,3), mirror_locations(:,1), mirror_locations(:,2),'--', 'Color', [0.8,0.8,0.8]);
    '''

    for element in optical_system:
        element.plot()

    beam = beam_injector.get_beam()
    for element in optical_system:
        with timer(f'Propagating {element}'):
            beam = element.propagate(beam)

    return beam




import sys

from numpy import array as arr
from numpy import pi, abs, cos, sin, rot90, linspace, ogrid, arange

import matplotlib.pyplot as plt

import tomli

from .beams.geometric import BeamGaussianGeometric
from .beam_injection import BeamInjection
from .elements.mirror import Mirror
from .elements.parabola import Parabola
from .elements.camera import Camera
from .propagate import propagate
from .elements import initialise_element

def parse_optical_system(config_struct):
    try:
        beam_injection = BeamInjection.from_dict(config_struct['beam_injection'])
    except KeyError:
        print('Error: no beam injector specified')
        sys.exit(1)
    try:
        element_list = config_struct['elements']
    except KeyError:
        print('Error: no elements specified')
        sys.exit(1)

    elements = [initialise_element(el) for el in element_list]
    
    return beam_injection, elements

try:
    with open('system.toml', 'rb') as config_file:
            config_struct = tomli.load(config_file)
except (FileNotFoundError, tomli.TOMLDecodeError):
    beam_injector = BeamInjection(
        position=arr([0,0,-1]),
        direction=arr([0,0,1]),
        wavelength=1e-6,
        beam_type=BeamGaussianGeometric,
        beam_waist=30e-3,
        distance_to_focus=0
    )

    f = 0.05
    Phi = pi/180 * 30
    f0 = f * cos(Phi)**2
    elements = [
        Mirror(arr([0, 0, 0]), [0, pi-pi/4], 5*25.4e-3),
        Mirror(arr([1, 0, 0]), [0,   -pi/4], 5*25.4e-3),
        Mirror(arr([1, 0, 1]), [0, pi+pi/4], 5*25.4e-3),
        Mirror(arr([0, 0, 1]), [0,   +pi/4], 5*25.4e-3),
        Parabola(arr([0, 0, 2]), [0.0, pi+0.0], radius=0.2, parent_focal_length=f0, parabola_angle=2*Phi),
        Camera(arr([f*sin(2*Phi), 0, 2-f*cos(2*Phi)]), [0, -2*Phi], (640, 480), arr([64, 48])*2e-3, 1e4)
    ]
    parabola = elements[-2]
    camera = elements[-1]
else:
    beam_injector, elements = parse_optical_system(config_struct)
    for el in elements:
        if isinstance(el, Camera):
            camera = el
        if isinstance(el, Parabola):
            parabola = el


fig, axs = plt.subplots(3, 3)
for i, dtheta in enumerate(arange(3)):
    parabola.pitch = 4e-5 * dtheta
    for j, dphi in enumerate(arange(3)):
        parabola.yaw = pi + 4e-5 * dphi
        print('starting propagation')
        propagate(beam_injector, elements)
        print('finished')
        axs[i,j].imshow(rot90(camera.read()), vmin=0, extent=(-32,32,-24,24))
        axs[i,j].axis('off')
        print(i,j)
        plt.imsave(f'images/phi_{dphi}_theta_{dtheta}.png', rot90(camera.read()), vmin=0, vmax=200)
#plt.show()





import numpy as np
from numpy import sqrt, sin, cos, tan, arctan2
from numpy import array as arr

from ..beams import Beam
from ..beams.geometric import BeamFresnel
from ..beams.hermite import BeamHermite
from ..beams.fresnel_gouy import BeamGouyFresnel
from .mirror import Mirror
from .. import util
from .. import parser

from matplotlib import pyplot as plt

class Parabola(Mirror):
    @staticmethod
    def from_dict(config_dict):
        position = parser.position(config_dict['position'])
        correct_angles = parser.euler_angles(config_dict['pointing'])
        try:
            radius = config_dict['radius']
        except KeyError:
            radius = float('inf')

        roll = 0
        match config_dict['off_axis']:
            case {'offset': [U0, V0]}:
                X0 = np.array([X0, Y0])
                Phi = None
                cosPhi2 = 1/(1 + X0**2 + Y0**2)
            case {'angle': value, **others}:
                Phi = parser.angle(value)
                X0 = None
                cosPhi2 = np.cos(Phi)**2
                if 'roll' in others:
                    roll = others['roll']

        match config_dict['focal_length']:
            case {'parent': value}:
                pfl = value
            case {'actual': value}:
                pfl = value * cosPhi2

        return Parabola(position, correct_angles, radius, pfl, Phi, roll, X0)

    def __init__(self, position, correct_angles, radius,
                 parent_focal_length, parabola_angle=None,
                 parabola_roll=0, X0=None):
        super().__init__(position, correct_angles, radius)
        self.f0 = parent_focal_length
        if parabola_angle is not None and X0 is not None:
            print('Parabola angle and X0 both specified. Using X0.')
        if X0 is not None:
            x0, y0 = X0[0]/(2*self.f0), X0[1]/(2*self.f0)
            self.X0 = np.array([x0, y0, (x0**2 + y0**2 - 1)/2])
        elif parabola_angle is not None:
            r0 = np.tan(parabola_angle/2)
            x0, y0 = r0 * np.cos(parabola_roll), r0 * np.sin(parabola_roll)
            self.X0 = np.array([x0, y0, (r0**2 - 1) / 2])

    def normal(self, hit):
        return np.array([-hit[0], -hit[1], 1]) / sqrt(1 + hit[0]**2 + hit[1]**2)

    def propagate(self, beam):
        transformed_beam = self.transform_beam(beam)
        hit_normal = self.normal(transformed_beam.position)
        reflect = util.reflection_matrix(hit_normal)
        new_beam_normal = reflect @ transformed_beam.normal

        x, y = np.mgrid[-10:10:1001j, -10:10:1001j]* beam.width()
        u = x / (2*self.f0)
        v = y / (2*self.f0)

        z_residual, focal_factor, clip_R = self.optical_path_difference(
            u, v, transformed_beam, False
        )

        nb_dot_nh = np.dot(hit_normal, -transformed_beam.normal)

        f = self.f0 / (focal_factor * nb_dot_nh)

        phase = beam.wavenumber * z_residual * nb_dot_nh
        mirror_e_field = beam.e_field(x, y) * np.exp(1j * phase)
        mirror_e_field[np.abs(clip_R.imag) > 0] = 0
        if beam.clipping:
            mirror_e_field *= (clip_R <= self.radius)
        newbeam = BeamGouyFresnel(
            position=transformed_beam.position,
            normal=new_beam_normal, wavenumber=beam.wavenumber,
            beam_width=beam.width(), u=x, v=y, focal_length=f,
            e_field=mirror_e_field
        )
        return self.parabola_to_world(newbeam)

    def world_to_parabola(self, beam):
        R_parabola = util.rot_matrix_angles(self.pitch, self.yaw).T
        parabola_position = self.X0 + R_parabola @ (
                beam.position - self.position
            ) / (2*self.f0)
        parabola_normal = R_parabola @ beam.normal
        return beam.copy_with(position=parabola_position, normal=parabola_normal)

    def parabola_to_world(self, beam):
        R_parabola = util.rot_matrix_angles(self.pitch, self.yaw)
        world_position = self.position + (
            (2*self.f0) * R_parabola @ (
                    beam.position - self.X0
            ))
        world_normal = R_parabola @ beam.normal
        return beam.copy_with(position=world_position, normal=world_normal)

    def transform_beam(self, beam):
        R_parabola = util.rot_matrix_angles(self.pitch, self.yaw).T
        transformed_position = self.X0 + R_parabola @ (
                beam.position - self.position
            ) / (2*self.f0)
        transformed_normal = R_parabola @ beam.normal
        x0, y0, z0 = transformed_position
        nx, ny, nz = transformed_normal

        if nz > 0:
            print('beam travelling in wrong direction')

        A = nx**2 + ny**2
        Bon2 = (nx * x0 + ny * y0) - nz
        C = x0**2 + y0**2 - 1 - 2*z0  # = 2 ( Z(x0, y0) - z0 )

        #A x^2 + 2 B x + C = 0
        #x = - C/(2 B)
        #- C^2/8B^3 = dx/dA

        if C > 0:
            print('beam behind parabola')

        param = A*C/Bon2**2
        delta = -C/(Bon2) / (sqrt(1 - param) + 1)

        hit_position = transformed_position + delta * transformed_normal

        dx, dy, dz = hit_position - self.X0
        if dx**2 + dy**2 > self.radius**2:
            print('missed mirror')

        return Beam(hit_position, transformed_normal, beam.wavenumber)

    def optical_path_difference(self, u, v, beam_ray, full_evaluation=True):
        '''
        Calculate optical path difference for specified beam with transverse coordinates u,v
        beam_ray: ray in parabola coordinates. position should already be in parabola coordinates,
            in contact with the parabola and in 'parabola reduced' coordinates, i.e. divided by 2f0
        u_beam, v_beam: coordinates to evaluate phases at in beam transverse directions.
        '''

        xh, yh, zh = beam_ray.position

        U, V, N = util.rot_matrix_normal(beam_ray.normal).T
        nx, ny, nz = N

        x0 = xh + U[0] * u + V[0] * v
        y0 = yh + U[1] * u + V[1] * v
        z0 = zh + U[2] * u + V[2] * v

        cosPhi = (1 + xh**2 + yh**2)**(-1/2)

        if nx**2 + ny**2 < 1e-20:
            hit_x = x0
            hit_y = y0

            OPD_aberration = 0
            focal_factor = cosPhi

        else:
            # delta (2nd order) = 1/2 ( cu2 u^2 + 2 cuv u v + cv2 v^2 )
            # = (cu2+cv2)/2 * (u^2+v^2)/2 + (cu2-cv2)/2 * (u^2-v^2)/2 + cuv u v
            # = { ----focusing part---- } + { --------astigmatism part------- }

            M = np.array([[1, 0, xh], [0, 1, yh], [xh, yh, xh**2 + yh**2]])
            A = np.array([-xh, -yh, 1])

            cu2 = (V @ M @ V) / (N @ A)**2
            cuv = (U @ M @ V) / (N @ A)**2
            cv2 = (U @ M @ U) / (N @ A)**2

            focal_factor = (cu2 + cv2)/2 * cosPhi
            focus = focal_factor * (u**2+v**2)/2

            if full_evaluation:
                A = nx**2 + ny**2
                Bon2 = (nx * x0 + ny * y0) - nz
                C = x0**2 + y0**2 - 1 - 2*z0

                param = A*C/Bon2**2
                delta = -C/(Bon2) / (sqrt(1 - param) + 1)

                hit_x = x0 + delta * nx
                hit_y = y0 + delta * ny

                Delta = ((hit_x - xh)**2 + (hit_y - yh)**2) / 2 * cosPhi

                OPD_aberration = Delta - focus
            else:
                dc = (cu2 - cv2)/2
                astigma_factor = sqrt(dc**2 + cuv**2)
                astigma_tilt = arctan2(-cuv, dc) / 2

                u_as = u * cos(astigma_tilt) + v * sin(astigma_tilt)
                v_as = v * cos(astigma_tilt) - u * sin(astigma_tilt)
                astigma = astigma_factor * ( u_as**2 - v_as**2 ) / 2

                delta0 = u * xh - v * yh - (u**2 + v**2) / 2
                
                hit_x = x0 + delta0 * nx
                hit_y = y0 + delta0 * ny
                
                coma = (u * nx - v * ny) * (u**2 + v**2) / 2
                OPD_aberration = ( astigma + coma ) * cosPhi

        clip_R = sqrt((hit_x - self.X0[0])**2 + (hit_y-self.X0[1])**2)

        return OPD_aberration, focal_factor, clip_R

